from urllib.parse import urljoin

import requests
from django.conf import settings
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, render
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import staff_member_required
from NEMO.models import MembershipHistory, Tool, User
from NEMO.views.users import get_identity_service


@staff_member_required
@require_GET
def qualifications(request):
	""" Present a web page to allow staff to qualify or disqualify users on particular tools. """
	users = User.objects.filter(is_active=True)
	tools = Tool.objects.filter(visible=True)
	return render(request, 'qualifications.html', {'users': users, 'tools': tools})


@staff_member_required
@require_POST
def modify_qualifications(request):
	""" Change the tools that a set of users is qualified to use. """
	action = request.POST.get('action')
	if action != 'qualify' and action != 'disqualify':
		return HttpResponseBadRequest("You must specify that you are qualifying or disqualifying users.")
	users = request.POST.getlist('chosen_user[]') or request.POST.get('chosen_user') or []
	users = User.objects.in_bulk(users)
	if users == {}:
		return HttpResponseBadRequest("You must specify at least one user.")
	tools = request.POST.getlist('chosen_tool[]') or request.POST.get('chosen_tool') or []
	tools = Tool.objects.in_bulk(tools)
	if tools == {}:
		return HttpResponseBadRequest("You must specify at least one tool.")

	for user in users.values():
		original_qualifications = set(user.qualifications.all())
		if action == 'qualify':
			user.qualifications.add(*tools)
			original_physical_access_levels = set(user.physical_access_levels.all())
			physical_access_level_automatic_enrollment = list(set([t.grant_physical_access_level_upon_qualification for t in tools.values() if t.grant_physical_access_level_upon_qualification]))
			user.physical_access_levels.add(*physical_access_level_automatic_enrollment)
			current_physical_access_levels = set(user.physical_access_levels.all())
			added_physical_access_levels = set(current_physical_access_levels) - set(original_physical_access_levels)
			for access_level in added_physical_access_levels:
				entry = MembershipHistory()
				entry.authorizer = request.user
				entry.parent_content_object = access_level
				entry.child_content_object = user
				entry.action = entry.Action.ADDED
				entry.save()
			if get_identity_service().get("available", False):
				for t in tools:
					tool = Tool.objects.get(id=t)
					if tool.grant_badge_reader_access_upon_qualification:
						parameters = {
							'username': user.username,
							'domain': user.domain,
							'requested_area': tool.grant_badge_reader_access_upon_qualification,
						}
						timeout = settings.IDENTITY_SERVICE.get('timeout', 3)
						requests.put(urljoin(settings.IDENTITY_SERVICE['url'], '/add/'), data=parameters, timeout=timeout)
		elif action == 'disqualify':
			user.qualifications.remove(*tools)
		current_qualifications = set(user.qualifications.all())
		# Record the qualification changes for each tool:
		added_qualifications = set(current_qualifications) - set(original_qualifications)
		for tool in added_qualifications:
			entry = MembershipHistory()
			entry.authorizer = request.user
			entry.parent_content_object = tool
			entry.child_content_object = user
			entry.action = entry.Action.ADDED
			entry.save()
		removed_qualifications = set(original_qualifications) - set(current_qualifications)
		for tool in removed_qualifications:
			entry = MembershipHistory()
			entry.authorizer = request.user
			entry.parent_content_object = tool
			entry.child_content_object = user
			entry.action = entry.Action.REMOVED
			entry.save()

	if request.POST.get('redirect') == 'true':
		dictionary = {
			'title': 'Success!',
			'heading': 'Success!',
			'content': 'Tool qualifications were successfully modified.',
		}
		return render(request, 'acknowledgement.html', dictionary)
	else:
		return HttpResponse()


@staff_member_required
@require_GET
def get_qualified_users(request):
	tool = get_object_or_404(Tool, id=request.GET.get('tool_id'))
	users = User.objects.filter(is_active=True)
	dictionary = {
		'tool': tool,
		'users': users,
		'expanded': True
	}
	return render(request, 'tool_control/qualified_users.html', dictionary)



from logging import getLogger
from re import search
from urllib.parse import urljoin

import requests
from django.conf import settings
from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.urls import reverse
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import staff_member_or_tool_superuser_required
from NEMO.exceptions import ProjectChargeException
from NEMO.models import MembershipHistory, Project, Tool, TrainingSession, User
from NEMO.views.policy import check_billing_to_project
from NEMO.views.users import get_identity_service

training_logger = getLogger(__name__)


@staff_member_or_tool_superuser_required
@require_GET
def training(request):
	""" Present a web page to allow staff or tool superusers to charge training and qualify users on particular tools. """
	user: User = request.user
	users = User.objects.filter(is_active=True).exclude(id=user.id)
	tools = Tool.objects.filter(visible=True)
	if not user.is_staff and user.is_tool_superuser:
		tools = tools.filter(_superusers__in=[user])
	return render(request, 'training/training.html', {'users': users, 'tools': tools, 'charge_types': TrainingSession.Type.Choices})


@staff_member_or_tool_superuser_required
@require_GET
def training_entry(request):
	entry_number = int(request.GET['entry_number'])
	return render(request, 'training/training_entry.html', {'entry_number': entry_number, 'charge_types': TrainingSession.Type.Choices})


def is_valid_field(field):
	return search("^(chosen_user|chosen_tool|chosen_project|duration|charge_type|qualify)__[0-9]+$", field) is not None


@staff_member_or_tool_superuser_required
@require_POST
def charge_training(request):
	trainer: User = request.user
	try:
		charges = {}
		for key, value in request.POST.items():
			if is_valid_field(key):
				attribute, separator, index = key.partition("__")
				index = int(index)
				if index not in charges:
					charges[index] = TrainingSession()
					charges[index].trainer = trainer
				if attribute == "chosen_user":
					charges[index].trainee = User.objects.get(id=to_int_or_negative(value))
				if attribute == "chosen_tool":
					charges[index].tool = Tool.objects.get(id=to_int_or_negative(value))
					if not trainer.is_staff and trainer.is_tool_superuser and charges[index].tool not in trainer.superuser_for_tools.all():
						raise Exception("The trainer is not authorized to train on this tool")
				if attribute == "chosen_project":
					charges[index].project = Project.objects.get(id=to_int_or_negative(value))
				if attribute == "duration":
					charges[index].duration = int(value)
				if attribute == "charge_type":
					charges[index].type = int(value)
				if attribute == "qualify":
					charges[index].qualified = (value == "on")
		for c in charges.values():
			c.full_clean()
			check_billing_to_project(c.project, c.trainee, c.tool)
	except ProjectChargeException as e:
		return HttpResponseBadRequest(e.msg)
	except User.DoesNotExist:
		return HttpResponseBadRequest("Please select a trainee from the list")
	except Tool.DoesNotExist:
		return HttpResponseBadRequest("Please select a tool from the list")
	except Project.DoesNotExist:
		return HttpResponseBadRequest("Please select a project from the list")
	except Exception as e:
		training_logger.exception(e)
		return HttpResponseBadRequest('An error occurred while processing the training charges. None of the charges were committed to the database. Please review the form for errors and omissions then submit the form again.')
	else:
		for c in charges.values():
			if c.qualified:
				qualify(c.trainer, c.trainee, c.tool)
			c.save()
		dictionary = {
			'title': 'Success!',
			'content': 'Training charges were successfully saved.',
			'redirect': reverse('landing'),
		}
		return render(request, 'display_success_and_redirect.html', dictionary)


def qualify(authorizer, user, tool):
	if tool in user.qualifications.all():
		return
	user.qualifications.add(tool)
	entry = MembershipHistory()
	entry.authorizer = authorizer
	entry.parent_content_object = tool
	entry.child_content_object = user
	entry.action = entry.Action.ADDED
	entry.save()

	if tool.grant_physical_access_level_upon_qualification:
		if tool.grant_physical_access_level_upon_qualification not in user.accessible_access_levels().all():
			user.physical_access_levels.add(tool.grant_physical_access_level_upon_qualification)
			entry = MembershipHistory()
			entry.authorizer = authorizer
			entry.parent_content_object = tool.grant_physical_access_level_upon_qualification
			entry.child_content_object = user
			entry.action = entry.Action.ADDED
			entry.save()

	if get_identity_service().get('available', False):
		if tool.grant_badge_reader_access_upon_qualification:
			parameters = {
				'username': user.username,
				'domain': user.domain,
				'requested_area': tool.grant_badge_reader_access_upon_qualification,
			}
			timeout = settings.IDENTITY_SERVICE.get('timeout', 3)
			requests.put(urljoin(settings.IDENTITY_SERVICE['url'], '/add/'), data=parameters, timeout=timeout)


def to_int_or_negative(value: str):
	try:
		return int(value)
	except ValueError:
		return -1



from django.forms import Widget
from django.urls import reverse
from django.utils.html import escape
from django.utils.safestring import mark_safe


class ConfigurationEditor(Widget):
	def __init__(self, attrs=None):
		self.url = reverse('tool_configuration')
		super().__init__(attrs)

	def render(self, name, value, attrs=None, **kwargs):
		result = ""
		for config in value["configurations"]:
			current_settings = config.current_settings_as_list()
			render_as_form = value.get("render_as_form", None)
			if render_as_form is None:
				render_as_form = not config.tool.in_use() and config.user_is_maintainer(value["user"])
			if len(current_settings) == 1:
				result += self.__render_for_one(config, render_as_form)
			else:
				result += self.__render_for_multiple(config, render_as_form)
		return mark_safe(result)

	def __render_for_one(self, config, render_as_form=None):
		current_setting = config.current_settings_as_list()[0]
		result = "<p><label class='form-inline'>" + escape(config.name) + ": "
		if render_as_form:
			result += "<select class='form-control' style='width:300px; max-width:100%' onchange=\"on_change_configuration('" + self.url +"', " + str(config.id) + ", 0, this.value)\">"
			for index, option in enumerate(config.available_settings_as_list()):
				result += "<option value=" + str(index)
				if option == current_setting:
					result += " selected"
				result += ">" + escape(option) + "</option>"
			result += "</select>"
		else:
			result += escape(current_setting)
		result += "</label></p>"
		return result

	def __render_for_multiple(self, config, render_as_form=None):
		result = "<p>" + escape(config.name) + ":<ul>"
		for setting_index, current_setting in enumerate(config.current_settings_as_list()):
			result += "<li>"
			if render_as_form:
				result += "<label class='form-inline'>" + escape(config.configurable_item_name) + " #" + str(setting_index + 1) + ": "
				result += "<select class='form-control' style='width:300px' onchange=\"on_change_configuration('" + self.url +"', " + str(config.id) + ", " + str(setting_index) + ", this.value)\">"
				for option_index, option in enumerate(config.available_settings_as_list()):
					result += "<option value=" + str(option_index)
					if option == current_setting:
						result += " selected"
					result += ">" + escape(option) + "</option>"
				result += "</select></label>"
			else:
				result += config.configurable_item_name + " #" + str(setting_index + 1) + ": " + escape(current_setting)
		result += "</ul></p>"
		return result



from django.core.management import BaseCommand

from NEMO.views.access_requests import send_email_weekend_access_notification


class Command(BaseCommand):
	help = (
		"Run every hour to trigger the email notification for weekend access."
		"Weekend access request email has to be set in customizations for this to work."
	)

	def handle(self, *args, **options):
		send_email_weekend_access_notification()



from typing import List, Dict

from django.db.models import QuerySet


class TreeItem:
	""" Helper class for tree items """

	item_type = None
	id: int = None
	item = None
	name: str = None
	tree_category: str = None
	category: str = None
	ancestors: List = []
	descendants: List = []
	children: List = []
	child_items: List = []
	is_leaf: bool = False
	is_root: bool = False

	def ancestor_ids(self, include_self=False):
		ids = [ancestor.id for ancestor in self.ancestors]
		if include_self:
			ids.append(self.id)
		return ids

	def __str__(self):
		return self.name


class ModelTreeHelper:
	"""
	Helper class for trees with models. Create a tree in memory with links to ancestors and descendants
	to help limit database queries.
	"""

	def __init__(self, model_class, parent_field: str = "parent", children_field: str = "children", only_fields=None):
		self.only_fields = ["name", "category"]
		if only_fields is not None:
			self.only_fields.extend(only_fields)
			self.only_fields = list(set(self.only_fields))
		query_set = type(model_class).objects.all().prefetch_related(children_field).only(*self.only_fields)
		self.roots: List = list(query_set.filter(**{f"{parent_field}__isnull": True}))
		self.leaves_queryset: QuerySet = query_set.filter(**{f"{children_field}__isnull": True})
		self.leaf_ids: List[int] = list(
			query_set.filter(**{f"{children_field}__isnull": True}).values_list("id", flat=True)
		)

		self.items: Dict[int, TreeItem] = {}
		self.build_tree(model_class, parent_field, children_field, [], None)
		# Second pass to populate descendants
		for key, item in self.items.items():
			if not item.is_leaf:
				res = []
				for child in item.child_items:
					res.append(self.items[child.id])
				item.children = res.copy()
				result = []
				for key1, item1 in self.items.items():
					if item in item1.ancestors:
						result.append(self.items[item1.id])
				item.descendants = result.copy()

	def build_tree(self, model_class, parent_field, children_field, ancestors: List[TreeItem], items=None):
		is_root = items is None
		if is_root:
			items = self.roots

		for item in items:
			tree_category = "/".join([ancestor.name for ancestor in ancestors])
			if item.category:
				tree_category += "/" + item.category if tree_category else item.category
			tree_item = TreeItem()
			tree_item.id = item.id
			tree_item.name = item.name
			tree_item.item = item
			tree_item.item_type = type(model_class)
			tree_item.tree_category = tree_category
			tree_item.category = item.category
			tree_item.ancestors = ancestors
			tree_item.is_root = is_root
			# Add only fields
			for field in self.only_fields:
				setattr(tree_item, field, getattr(item, field))
			if item.id in self.leaf_ids:
				tree_item.is_leaf = True
			else:
				children = list(
					type(model_class)
						.objects.filter(**{f"{parent_field}__id": item.id})
						.prefetch_related(children_field)
						.only(*self.only_fields)
				)
				tree_item.child_items = children
				tree_item.is_leaf = False
				new_ancestors = ancestors.copy()
				new_ancestors.append(tree_item)
				self.build_tree(model_class, parent_field, children_field, new_ancestors, children)
			self.items[item.id] = tree_item

	def get_areas(self, ids: List[int]) -> List[TreeItem]:
		return [self.items[pk] for pk in ids]

	def get_ancestor_areas(self, tree_items: List[TreeItem], include_self=False):
		return self.get_areas(list(set([pk for tree_item in tree_items for pk in tree_item.ancestor_ids(include_self)])))

	def get_area(self, pk: int) -> TreeItem:
		return self.items.get(pk, None)


def get_area_model_tree():
	from NEMO.models import Area

	only_fields = ["name", "category", "maximum_capacity", "reservation_warning", "count_staff_in_occupancy", "count_service_personnel_in_occupancy"]
	return ModelTreeHelper(Area(), "parent_area", "area_children_set", only_fields)



from django.core.management import BaseCommand

from NEMO.views.calendar import do_create_closure_alerts


class Command(BaseCommand):
	help = (
		"Run every day to trigger the automatic alert creation for facility closures."
		"The field 'alert_days_before' has to be set on a Closure object for this to work."
	)

	def handle(self, *args, **options):
		do_create_closure_alerts()



# Generated by Django 2.2.13 on 2020-12-10 22:06

import NEMO.utilities
from django.db import migrations, models

from NEMO.migrations_utils import create_news_for_version


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0027_version_3_8_0'),
    ]

    def new_version_news(apps, schema_editor):
        create_news_for_version(apps, "3.9.0")

    operations = [
        migrations.AlterModelOptions(
            name='toolusagecounter',
            options={'ordering': ['tool__name']},
        ),
        migrations.AlterField(
            model_name='toolusagecounter',
            name='last_reset_value',
            field=models.FloatField(blank=True, help_text='The last value before the counter was reset', null=True),
        ),
        migrations.AlterField(
            model_name='toolusagecounter',
            name='value',
            field=models.FloatField(default=0, help_text='The current value of this counter'),
        ),
        migrations.CreateModel(
            name='ToolDocuments',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('document', models.FileField(null=True, blank=True, upload_to=NEMO.utilities.get_tool_document_filename, verbose_name='Document')),
                ('uploaded_at', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(blank=True, help_text='The optional name to display for this document', max_length=200, null=True)),
                ('url', models.CharField(blank=True, max_length=200, null=True, verbose_name='URL')),
                ('tool', models.ForeignKey(on_delete=models.deletion.CASCADE, to='NEMO.Tool')),
            ],
            options={
                'verbose_name_plural': 'Tool documents',
                'ordering': ['-uploaded_at'],
            },
        ),
        migrations.AddField(
            model_name='user',
            name='managed_projects',
            field=models.ManyToManyField(blank=True, help_text='Select the projects that this user is a PI for.', related_name='manager_set', to='NEMO.Project'),
        ),
        migrations.RunPython(new_version_news),
    ]



# Generated by Django 2.2.10 on 2020-02-17 16:59

from django.db import migrations, models
import NEMO.utilities

class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0012_version_2_0_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='interlockcard',
            name='name',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='tool',
            name='_description',
            field=models.TextField(blank=True, db_column='description', help_text='HTML syntax could be used', null=True),
        ),
        migrations.AddField(
            model_name='tool',
            name='_serial',
            field=models.CharField(blank=True, db_column='serial', help_text='Serial Number', max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='tool',
            name='_image',
            field=models.ImageField(blank=True, db_column='image', help_text='An image that represent the tool. Maximum width and height are 500px', upload_to=NEMO.utilities.get_tool_image_filename),
        ),
        migrations.AddField(
            model_name='physicalaccesslevel',
            name='allow_staff_access',
            field=models.BooleanField(default=False, help_text='Check this box to allow access to Staff users without explicitly granting them access'),
        ),
        migrations.CreateModel(
            name='AlertCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
            ],
            options={
                'verbose_name_plural': 'Alert categories',
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='alert',
            name='category',
            field=models.CharField(blank=True, help_text='A category/type for this alert.', max_length=200),
        ),
        migrations.AddField(
            model_name='alert',
            name='deleted',
            field=models.BooleanField(default=False, help_text="Indicates the alert has been deleted and won't be shown anymore"),
        ),
        migrations.AddField(
            model_name='alert',
            name='expired',
            field=models.BooleanField(default=False, help_text="Indicates the alert has expired and won't be shown anymore"),
        ),
    ]



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

from NEMO.migrations_utils import create_news_for_version


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0034_version_3_13_0'),
    ]

    def new_version_news(apps, schema_editor):
        create_news_for_version(apps, "3.14.0")

    def rename_customization(apps, schema_editor):
        Customization = apps.get_model("NEMO", "Customization")
        try:
            buddy_board_desc_old = Customization.objects.get(name="buddy_board_disclaimer")
            Customization.objects.create(name="buddy_board_description", value=buddy_board_desc_old.value)
            buddy_board_desc_old.delete()
        except:
            pass

    operations = [
        migrations.RunPython(new_version_news),
        migrations.RunPython(rename_customization),
        migrations.CreateModel(
            name='TemporaryPhysicalAccess',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_time', models.DateTimeField(help_text='The start of the temporary access')),
                ('end_time', models.DateTimeField(help_text='The end of the temporary access')),
                ('physical_access_level', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.PhysicalAccessLevel')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={'ordering': ['-end_time']},
        ),
        migrations.AddField(
            model_name='user',
            name='is_facility_manager',
            field=models.BooleanField(default=False, help_text='Designates this user as facility manager. Facility managers receive updates on all reported problems in the facility and can also review access requests.', verbose_name='facility manager'),
        ),
        migrations.AlterField(
            model_name='user',
            name='is_staff',
            field=models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff'),
        ),
        migrations.AlterField(
            model_name='user',
            name='is_superuser',
            field=models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='administrator'),
        ),
        migrations.AlterField(
            model_name='user',
            name='is_technician',
            field=models.BooleanField(default=False, help_text='Specifies how to bill staff time for this user. When checked, customers are billed at technician rates.', verbose_name='technician'),
        ),
        migrations.AlterField(
            model_name='landingpagechoice',
            name='notifications',
            field=models.CharField(blank=True, choices=[('news', 'News creation and updates - notifies all users'), ('safetyissue', 'New safety issues - notifies staff only'), ('buddyrequest', 'New buddy request - notifies all users'), ('buddyrequestmessage', 'New buddy request reply - notifies request creator and users who have replied'), ('temporaryphysicalaccessrequest', 'New access request - notifies other users on request and reviewers')], help_text="Displays a the number of new notifications for the user. For example, if the user has two unread news notifications then the number '2' would appear for the news icon on the landing page.", max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='emaillog',
            name='category',
            field=models.IntegerField(choices=[(0, 'General'), (1, 'System'), (2, 'Direct Contact'), (3, 'Broadcast Email'), (4, 'Timed Services'), (5, 'Feedback'), (6, 'Abuse'), (7, 'Safety'), (8, 'Tasks'), (9, 'Access Requests')], default=0),
        ),
        migrations.AlterField(
            model_name='physicalaccesslevel',
            name='schedule',
            field=models.IntegerField(choices=[(0, 'Anytime'), (1, 'Weekdays'), (2, 'Weekends')]),
        ),
        migrations.AddField(
            model_name='physicalaccesslevel',
            name='allow_user_request',
            field=models.BooleanField(default=False, help_text='Check this box to allow users to request this access temporarily in "Access requests"'),
        ),
        migrations.CreateModel(
            name='TemporaryPhysicalAccessRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_time', models.DateTimeField(auto_now_add=True, help_text='The date and time when the request was created.')),
                ('last_updated', models.DateTimeField(auto_now=True, help_text='The last time this request was modified.')),
                ('description', models.TextField(blank=True, help_text='The description of the request.', null=True)),
                ('start_time', models.DateTimeField(help_text='The requested time for the access to start.')),
                ('end_time', models.DateTimeField(help_text='The requested time for the access to end.')),
                ('status', models.IntegerField(choices=[(0, 'Pending'), (1, 'Approved'), (2, 'Denied'), (3, 'Expired')], default=0)),
                ('deleted', models.BooleanField(default=False, help_text="Indicates the request has been deleted and won't be shown anymore.")),
                ('creator', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='access_requests_created', to=settings.AUTH_USER_MODEL)),
                ('last_updated_by', models.ForeignKey(blank=True, help_text='The last user who modified this request.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='access_requests_updated', to=settings.AUTH_USER_MODEL)),
                ('other_users', models.ManyToManyField(blank=True, help_text='Select the other users requesting access.', to=settings.AUTH_USER_MODEL)),
                ('physical_access_level', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.PhysicalAccessLevel')),
                ('reviewer', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='access_requests_reviewed', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-creation_time'],
            },
        ),
        migrations.AddField(
            model_name='staffcharge',
            name='note',
            field=models.TextField(blank=True, null=True),
        ),
    ]
